//****************************************************************************
// Copyright (C) 2001,2002  PEAK System-Technik GmbH
//
// linux@peak-system.com
// www.peak-system.com 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// Maintainer(s): Klaus Hitschler (klaus.hitschler@gmx.de)
//****************************************************************************

//****************************************************************************
//
// pcan_sja1000.c - all about sja1000 init and data handling
//
// $Log: pcan_sja1000.c,v $
// Revision 1.22  2002/02/11 18:26:32  klaus
// moved libpcan.h and pcan.h to new locations
//
// Revision 1.21  2002/01/30 20:54:27  klaus
// simple source file header change
//
// Revision 1.20  2002/01/21 21:00:59  klaus
// Fifo buffer allocation write fault removed
//
// Revision 1.19  2002/01/20 16:26:53  klaus
// Removed bug in interrupt handling of PCI, reorderd req_irq call
//
// Revision 1.18  2002/01/13 18:36:12  klaus
// first release
//
// Revision 1.17  2002/01/05 16:19:25  klaus
// added extended frame rejection
//
//****************************************************************************

//****************************************************************************
// INCLUDES
#include <src/pcan_common.h>
#include <linux/sched.h>
#include <asm/errno.h>
#include <asm/byteorder.h>  // because of little / big endian
#include <linux/delay.h>
#include <src/pcan_main.h>
#include <pcan.h>
#include <src/pcan_fifo.h>

//****************************************************************************
// DEFINES

// sja1000 registers, only PELICAN mode - TUX like it
#define MODE                   0      // mode register
#define COMMAND                1
#define CHIPSTATUS             2
#define INTERRUPT_STATUS       3
#define INTERRUPT_ENABLE       4      // acceptance code
#define TIMING0                6      // bus timing 0
#define TIMING1                7      // bus timing 1
#define OUTPUT_CONTROL         8      // output control
#define TESTREG                9

#define ARBIT_LOST_CAPTURE    11      // transmit buffer: Identifier
#define ERROR_CODE_CAPTURE    12      // RTR bit und data length code
#define ERROR_WARNING_LIMIT   13      // start byte of data field
#define RX_ERROR_COUNTER      14
#define TX_ERROR_COUNTER      15

#define ACCEPTANCE_CODE_BASE  16
#define RECEIVE_FRAME_BASE    16
#define TRANSMIT_FRAME_BASE   16

#define ACCEPTANCE_MASK_BASE  20

#define RECEIVE_MSG_COUNTER   29
#define RECEIVE_START_ADDRESS 30

#define CLKDIVIDER            31      // set bit rate and pelican mode

// important sja1000 register contents, MODE register
#define SLEEP_MODE             0x10
#define ACCEPT_FILTER_MODE     0x08
#define SELF_TEST_MODE         0x04
#define LISTEN_ONLY_MODE       0x02
#define RESET_MODE             0x01
#define NORMAL_MODE            0x00

// COMMAND register
#define CLEAR_DATA_OVERRUN     0x08
#define RELEASE_RECEIVE_BUFFER 0x04
#define ABORT_TRANSMISSION     0x02
#define TRANSMISSION_REQUEST   0x01

// CHIPSTATUS register
#define BUS_STATUS             0x80
#define ERROR_STATUS           0x40
#define TRANSMIT_STATUS        0x20
#define RECEIVE_STATUS         0x10
#define TRANS_COMPLETE_STATUS  0x08
#define TRANS_BUFFER_STATUS    0x04
#define DATA_OVERRUN_STATUS    0x02
#define RECEIVE_BUFFER_STATUS  0x01

// INTERRUPT STATUS register
#define BUS_ERROR_INTERRUPT    0x80
#define ARBIT_LOST_INTERRUPT   0x40
#define ERROR_PASSIV_INTERRUPT 0x20
#define WAKE_UP_INTERRUPT      0x10
#define DATA_OVERRUN_INTERRUPT 0x08
#define ERROR_WARN_INTERRUPT   0x04
#define TRANSMIT_INTERRUPT     0x02
#define RECEIVE_INTERRUPT      0x01

// INTERRUPT ENABLE register
#define BUS_ERROR_INTERRUPT_ENABLE    0x80
#define ARBIT_LOST_INTERRUPT_ENABLE   0x40
#define ERROR_PASSIV_INTERRUPT_ENABLE 0x20
#define WAKE_UP_INTERRUPT_ENABLE      0x10
#define DATA_OVERRUN_INTERRUPT_ENABLE 0x08
#define ERROR_WARN_INTERRUPT_ENABLE   0x04
#define TRANSMIT_INTERRUPT_ENABLE     0x02
#define RECEIVE_INTERRUPT_ENABLE      0x01

// OUTPUT CONTROL register
#define OUTPUT_CONTROL_TRANSISTOR_P1  0x80
#define OUTPUT_CONTROL_TRANSISTOR_N1  0x40
#define OUTPUT_CONTROL_POLARITY_1     0x20
#define OUTPUT_CONTROL_TRANSISTOR_P0  0x10
#define OUTPUT_CONTROL_TRANSISTOR_N0  0x08
#define OUTPUT_CONTROL_POLARITY_0     0x04
#define OUTPUT_CONTROL_MODE_1         0x02
#define OUTPUT_CONTROL_MODE_0         0x01

// TRANSMIT or RECEIVE BUFFER
#define BUFFER_EFF                    0x80 // set for 29 bit identifier
#define BUFFER_RTR                    0x40 // set for RTR request
#define BUFFER_DLC_MASK               0x0f

// CLKDIVIDER register
#define CAN_MODE                      0x80
#define CAN_BYPASS                    0x40
#define RXINT_OUTPUT_ENABLE           0x20
#define CLOCK_OFF                     0x08
#define CLOCK_DIVIDER_MASK            0x07

// additional informations
#define CLOCK_HZ                  16000000 // crystal frequency

// time for mode register to change mode
#define MODE_REGISTER_SWITCH_TIME 100 // msec 

// some CLKDIVIDER register contents, hardware architecture dependend 
#define PELICAN_SINGLE  (CAN_MODE | CAN_BYPASS | 0x07 | CLOCK_OFF)
#define PELICAN_MASTER  (CAN_MODE | CAN_BYPASS | 0x07            )
#define PELICAN_DEFAULT (CAN_MODE                                )
#define CHIP_RESET      PELICAN_SINGLE 

// hardware depended setup for OUTPUT_CONTROL register
#define OUTPUT_CONTROL_SETUP (OUTPUT_CONTROL_TRANSISTOR_P0 | OUTPUT_CONTROL_TRANSISTOR_N0 | OUTPUT_CONTROL_MODE_1)

// the interrupt enables
#define INTERRUPT_ENABLE_SETUP (RECEIVE_INTERRUPT_ENABLE | TRANSMIT_INTERRUPT_ENABLE | DATA_OVERRUN_INTERRUPT_ENABLE | BUS_ERROR_INTERRUPT_ENABLE | ERROR_PASSIV_INTERRUPT_ENABLE)

// the maximum number of handled messages in one interrupt 
#define MAX_MESSAGES_PER_INTERRUPT 8

// the maximum number of handled sja1000 interrupts in 1 handler entry
#define MAX_INTERRUPTS_PER_ENTRY   4

// a helper for fast conversion between SJA1000 data ordering and host data order
typedef union
{
  u8  ucID[4];
  u32 dwID;
} ID;

//****************************************************************************
// GLOBALS

//****************************************************************************
// LOCALS

//****************************************************************************
// CODE  

//----------------------------------------------------------------------------
// switches the chip into reset mode
static int set_reset_mode(struct pcandev *dev)
{
  u32 dwStart = get_mtime();
  u8  tmp;
  
  tmp = dev->readreg(dev, MODE);
  while (!(tmp & RESET_MODE) && ((get_mtime() - dwStart) < MODE_REGISTER_SWITCH_TIME)) 
  {
    dev->writereg(dev, MODE, RESET_MODE); // force into reset mode
    wmb();
    schedule();
    tmp = dev->readreg(dev, MODE);
  }   
  
  if (!(tmp & RESET_MODE))
    return -EIO;  
  else
    return 0;  
} 

//----------------------------------------------------------------------------
// switches the chip back from reset mode
static int set_normal_mode(struct pcandev *dev, u8 ucModifier)
{
  u32 dwStart = get_mtime();
  u8  tmp;
  
  tmp = dev->readreg(dev, MODE);
  while ((tmp != ucModifier) && ((get_mtime() - dwStart) < MODE_REGISTER_SWITCH_TIME))
  {
    dev->writereg(dev, MODE, ucModifier); // force into normal mode
    wmb();
    schedule();
    tmp = dev->readreg(dev, MODE);
  }   
  
  if (tmp != ucModifier)
    return -EIO;  
  else
    return 0;  
} 

//----------------------------------------------------------------------------
// init CAN-chip
int sja1000_open(struct pcandev *dev, u16 btr0btr1, u8 bExtended, u8 bListenOnly)
{
  int result     = 0;
  u8  clkdivider = PELICAN_DEFAULT;
  u8  ucModifier = (bListenOnly) ? LISTEN_ONLY_MODE : NORMAL_MODE;
  
  DPRINTK(KERN_DEBUG "%s: sja1000_open()\n", DEVICE_NAME);
  
  // switch to reset 
  result = set_reset_mode(dev);
  if (result)
    goto fail;    
    
  // take a fresh status
  dev->wCANStatus = 0;
  
  // store extended mode (standard still accepted)
  dev->bExtended = bExtended;
    
  // configure clock divider register, switch into pelican mode, depended of of type
  if (dev->wType == HW_PCI)
    clkdivider = (dev->port.pci.ucMasterDevice == CHANNEL_MASTER) ? PELICAN_MASTER : PELICAN_SINGLE;
  
  dev->writereg(dev, CLKDIVIDER, clkdivider);
  
  // configure acceptance code registers
  dev->writereg(dev, ACCEPTANCE_CODE_BASE,     0);
  dev->writereg(dev, ACCEPTANCE_CODE_BASE + 1, 0);
  dev->writereg(dev, ACCEPTANCE_CODE_BASE + 2, 0);
  dev->writereg(dev, ACCEPTANCE_CODE_BASE + 3, 0);
  
  // configure all acceptance mask registers to don't care 
  dev->writereg(dev, ACCEPTANCE_MASK_BASE,     0xff);
  dev->writereg(dev, ACCEPTANCE_MASK_BASE + 1, 0xff);
  dev->writereg(dev, ACCEPTANCE_MASK_BASE + 2, 0xff);
  dev->writereg(dev, ACCEPTANCE_MASK_BASE + 3, 0xff);
  
  // configure bus timing registers
  dev->writereg(dev, TIMING0, (u8)((btr0btr1 >> 8) & 0xff));
  dev->writereg(dev, TIMING1, (u8)((btr0btr1     ) & 0xff));  
  
  // configure output control registers
  dev->writereg(dev, OUTPUT_CONTROL, OUTPUT_CONTROL_SETUP);
  
  // clear any pending interrupt
  dev->readreg(dev, INTERRUPT_STATUS);
  
  // enter normal operating mode
  result = set_normal_mode(dev, ucModifier);
  if (result)
    goto fail;    
  
  // enable CAN interrupts
  dev->writereg(dev, INTERRUPT_ENABLE, INTERRUPT_ENABLE_SETUP);
  
  fail:
  return result;
}

//----------------------------------------------------------------------------
// release CAN-chip
void sja1000_release(struct pcandev *dev) 
{
  DPRINTK(KERN_DEBUG "%s: sja1000_release()\n", DEVICE_NAME);
  
  // abort pending transmissions
  dev->writereg(dev, COMMAND, ABORT_TRANSMISSION);
    
  // disable CAN interrupts and set chip in reset mode
  dev->writereg(dev, INTERRUPT_ENABLE, 0);
  set_reset_mode(dev);
}

//----------------------------------------------------------------------------
// read CAN-data from chip, supposed a message is available
static int sja1000_read(struct pcandev *dev)
{
  int i = MAX_MESSAGES_PER_INTERRUPT;
  int	result   = 0;
  TPCANRdMsg   *m;
  u8           rxfi;  // receive frame information byte
  u8           ucLen;
  ID           local;
  int          j;

  DPRINTK(KERN_DEBUG "%s: sja1000_read() %d\n", DEVICE_NAME, dev->readFifo.nStored);
  
  do
  {
    // aquire a fifo element
    result = pcan_fifo_claim_for_put(&dev->readFifo, (void *)&m);
    if (result)
      goto fail;
        
    // fix time of reading   
    m->dwTime  = get_mtime() - *dev->pdwInitTime;
    
    rxfi       = dev->readreg(dev, RECEIVE_FRAME_BASE);
    
    // filter out extended messages in non extended mode
    if (dev->bExtended || !(rxfi & BUFFER_EFF))
    { 
      m->Msg.LEN = ucLen = rxfi & BUFFER_DLC_MASK;    
      if (ucLen > 8)
        ucLen = 8;
  
      j = 0;
      if (rxfi & BUFFER_EFF)
      {
        m->Msg.MSGTYPE = MSGTYPE_EXTENDED;
  
        #ifdef __LITTLE_ENDIAN
        local.ucID[3] = dev->readreg(dev, RECEIVE_FRAME_BASE + 1);
        local.ucID[2] = dev->readreg(dev, RECEIVE_FRAME_BASE + 2);
        local.ucID[1] = dev->readreg(dev, RECEIVE_FRAME_BASE + 3);
        local.ucID[0] = dev->readreg(dev, RECEIVE_FRAME_BASE + 4);
        #else
        local.ucID[0] = dev->readreg(dev, RECEIVE_FRAME_BASE + 1);
        local.ucID[1] = dev->readreg(dev, RECEIVE_FRAME_BASE + 2);
        local.ucID[2] = dev->readreg(dev, RECEIVE_FRAME_BASE + 3);
        local.ucID[3] = dev->readreg(dev, RECEIVE_FRAME_BASE + 4);
        #endif
  
        local.dwID   >>= 3;

        while (ucLen--)
        {
          m->Msg.DATA[j] = dev->readreg(dev, RECEIVE_FRAME_BASE + 5 + j);
	        j++;
	      }
      }
      else
      {
        m->Msg.MSGTYPE = MSGTYPE_STANDARD;
  
        #ifdef __LITTLE_ENDIAN
        local.ucID[3] = dev->readreg(dev, RECEIVE_FRAME_BASE + 1);
        local.ucID[2] = dev->readreg(dev, RECEIVE_FRAME_BASE + 2);
        #else
        local.ucID[0] = dev->readreg(dev, RECEIVE_FRAME_BASE + 1);
        local.ucID[1] = dev->readreg(dev, RECEIVE_FRAME_BASE + 2);
        #endif
  
        local.dwID   >>= 21;

        while (ucLen--)
        {
          m->Msg.DATA[j] = dev->readreg(dev, RECEIVE_FRAME_BASE + 3 + j);
	        j++;
	      }
      }

      m->Msg.ID = local.dwID;

      if (rxfi & BUFFER_RTR)
        m->Msg.MSGTYPE |= MSGTYPE_RTR;
    
      // step forward in fifo
      result = pcan_fifo_put(&dev->readFifo);
      if (result)
        goto fail;  
	  }
	  else
	  {
	    // filtered out extended messages in non extended mode, reject claim
      result = pcan_fifo_put_reject(&dev->readFifo);
      if (result)
        goto fail; 
	  }     
   
    // release the receive buffer
    dev->writereg(dev, COMMAND, RELEASE_RECEIVE_BUFFER);  
    wmb();
    
    // give time to settle
    udelay(1);
	}	while (dev->readreg(dev, CHIPSTATUS) & RECEIVE_BUFFER_STATUS && (i--));	
      
  fail:
  return result ;
}

//----------------------------------------------------------------------------
// write CAN-data to chip, 
int sja1000_write(struct pcandev *dev) 
{
  int        result;
  int        j;
  TPCANMsg   *m;
  u8         txfi;
  ID         local;
  u8         ucLen;
  
  DPRINTK(KERN_DEBUG "%s: sja1000_write() %d\n", DEVICE_NAME, dev->writeFifo.nStored);
  
  // aquire a fifo element
  result = pcan_fifo_claim_for_get(&dev->writeFifo, (void *)&m);
  if (result)
    goto fail;
  
  txfi = ucLen = m->LEN;    
  
  j = 0;
  local.dwID = m->ID;
  if (m->MSGTYPE & MSGTYPE_EXTENDED)
  {
    local.dwID   <<= 3;

    #ifdef __LITTLE_ENDIAN
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 1, local.ucID[3]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 2, local.ucID[2]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 3, local.ucID[1]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 4, local.ucID[0]);
    #else
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 1, local.ucID[0]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 2, local.ucID[1]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 3, local.ucID[2]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 4, local.ucID[3]);
    #endif
  
    while (ucLen--)
    {
      dev->writereg(dev, TRANSMIT_FRAME_BASE + 5 + j, m->DATA[j]);
	    j++;
	  }
	
    txfi |= BUFFER_EFF;
  }
  else
  {
    local.dwID   <<= 21;
  
    #ifdef __LITTLE_ENDIAN
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 1, local.ucID[3]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 2, local.ucID[2]);
    #else
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 1, local.ucID[0]);
    dev->writereg(dev, TRANSMIT_FRAME_BASE + 2, local.ucID[1]);
    #endif
  

    while (ucLen--)
    {
      dev->writereg(dev, TRANSMIT_FRAME_BASE + 3 + j, m->DATA[j]);
	    j++;
	  }
  }

  if (m->MSGTYPE & MSGTYPE_RTR)
    txfi |= BUFFER_RTR;
  
  // release fifo buffer and step forward in fifo
  result = pcan_fifo_get(&dev->writeFifo);
  if (result)
    goto fail;
 
  // finish message construction
  dev->writereg(dev, TRANSMIT_FRAME_BASE, txfi); 

  // request a transmission
 	dev->writereg(dev, COMMAND, TRANSMISSION_REQUEST); 
	wmb(); 
	
  fail:
  return result;
}

//----------------------------------------------------------------------------
// handle a interrupt request
void sja1000_irqhandler(int irq, void *dev_id, struct pt_regs *regs)
{
  register struct pcandev *dev = (struct pcandev *)dev_id;
  u8 status;
  int err;
  u16 rwakeup = 0;
  u16 wwakeup = 0;
  int j = MAX_INTERRUPTS_PER_ENTRY;
  
   
  while ((j--) && (status = dev->readreg(dev, INTERRUPT_STATUS)))
  {    
    DPRINTK(KERN_DEBUG "%s: sja1000_irqhandler(0x%02x)\n", DEVICE_NAME, status);
  
    dev->dwInterruptCounter++;

    if (status & TRANSMIT_INTERRUPT)
    {
      // handle transmission
	    if ((err = sja1000_write(dev)))
	    {
	      if (err == -ENODATA)
		      wwakeup++;
		    else
		    {
	        dev->nLastError = err;
		      dev->dwErrorCounter++; 
		      dev->wCANStatus |= CAN_ERR_QXMTFULL; // fatal error!
		    }
		  }
    }
  
    if (status & DATA_OVERRUN_INTERRUPT)
    {
      // handle data overrun
	    dev->wCANStatus |= CAN_ERR_QOVERRUN;
      rwakeup++;
		  dev->dwErrorCounter++;

	    dev->writereg(dev, COMMAND, CLEAR_DATA_OVERRUN);  
      wmb();
    }   
    
    if (status & RECEIVE_INTERRUPT)
    {
      // handle receiption
	    if ((err = sja1000_read(dev)))
	    {
	      dev->nLastError = err;
		    dev->dwErrorCounter++;
	      dev->wCANStatus |=  CAN_ERR_QOVERRUN;

	      // throw away last message which was refused by fifo
        dev->writereg(dev, COMMAND, RELEASE_RECEIVE_BUFFER);  
        wmb();
		  }
		  rwakeup++;
    }
  
    if (status & (BUS_ERROR_INTERRUPT | ERROR_PASSIV_INTERRUPT))
    {
		  if (status & ERROR_PASSIV_INTERRUPT)
		  {
		    dev->wCANStatus |=  CAN_ERR_BUSHEAVY;
		    dev->dwErrorCounter++;		    
      }
		   
			// don't restart sja1000 after Bus Off
			if (status & BUS_ERROR_INTERRUPT)
			{
			  dev->wCANStatus |=  CAN_ERR_BUSOFF;
		    dev->dwErrorCounter++;		    
			}
    } 
  }
  
  if (wwakeup)
  {
    atomic_set(&dev->DataSendReady, 1); // signal to write I'm ready
    wake_up_interruptible(&dev->write_queue);
  }
  
  if (rwakeup)
    wake_up_interruptible(&dev->read_queue);
}

//----------------------------------------------------------------------------
// probe for a sja1000 - use it only in reset mode!
int  sja1000_probe(struct pcandev *dev)
{
  u8  tmp;
  u8  clkdivider;
  
  DPRINTK(KERN_DEBUG "%s: sja1000_probe()\n", DEVICE_NAME);
  
  // clkdivider is depended of of type
  switch (dev->wType)
  {
    case HW_PCI:           
      clkdivider = (dev->port.pci.ucMasterDevice == CHANNEL_MASTER) ? PELICAN_MASTER : PELICAN_SINGLE;
      break;
      
    default:
      clkdivider = PELICAN_DEFAULT;
      
      // make a hardware reset
      dev->writereg(dev, CLKDIVIDER, CHIP_RESET); 
      break;
  }
  
  // trace the clockdivider register to test for sja1000 / 82c200
	tmp = dev->readreg(dev, CLKDIVIDER);
  DPRINTK(KERN_DEBUG "%s: CLKDIVIDER traced (0x%02x)\n", DEVICE_NAME, tmp); 
	if (tmp & 0x10)
	  goto fail;
	  						  		  
  // until here, it's either a 82c200 or a sja1000
  if (set_reset_mode(dev))
    goto fail;    

  dev->writereg(dev, CLKDIVIDER, clkdivider); // switch to PeliCAN mode 
  dev->writereg(dev, INTERRUPT_ENABLE, 0);    // prospective clear all interrupts
  wmb();   
  DPRINTK(KERN_DEBUG "%s: Hopefully switched to PeliCAN mode\n", DEVICE_NAME);
  
  tmp = dev->readreg(dev, CHIPSTATUS);
  DPRINTK(KERN_DEBUG "%s: CHIPSTATUS traced (0x%02x)\n", DEVICE_NAME, tmp); 
  if ((tmp & 0x30) != 0x30)
    goto fail;
  
  tmp = dev->readreg(dev, INTERRUPT_STATUS);
  DPRINTK(KERN_DEBUG "%s: INTERRUPT_STATUS traced (0x%02x)\n", DEVICE_NAME, tmp);
  if (tmp & 0xfb)
    goto fail;
 
  tmp = dev->readreg(dev, RECEIVE_MSG_COUNTER);
  DPRINTK(KERN_DEBUG "%s: RECEIVE_MSG_COUNTER traced (0x%02x)\n", DEVICE_NAME, tmp);
  if (tmp)
    goto fail;
    
  DPRINTK(KERN_DEBUG "%s: sja1000_probe() is OK\n", DEVICE_NAME);
  return 0;
    
  fail:
  DPRINTK(KERN_DEBUG "%s: sja1000_probe() failed\n", DEVICE_NAME);
  return -ENXIO; // no such device or address
}

// end


